#! /usr/bin/env python
'''
The ``compile`` module converts instructions from pythonic data structures
into flat lists.abs

These may require further processing before being passed to the
``enact`` module for action.
'''
import six

from pyqubes.utils import flatten_list


def flags_boolean(flags):
    '''
    Return a list of string values, corresponding to the given keys whose values evaluate to True

    All keys and values will be converted to strings.

    :param dict flags: A dictionary in the form ``{'--flag': boolean}``, where ``boolean`` is used
            to determine whether ``--flag`` is included in the output.
    :returns: A flat list of strings
    '''
    flags_list = [str(k) for k, v in six.iteritems(flags) if v]
    return flags_list

def flags_store(flags):
    '''
    Return a list of string values, corresponding to the given keys and values whose values evaluate to True
    
    The output is a flat list of all strings.
    
    All keys and values will be converted to strings.

    :param dict flags: A dictionary in the form ``{'--flag': value}``, where ``value`` is used
            to determine whether the entry is included in the output.
    :returns: A flat list of strings
    '''
    flags_nested = [[str(k), str(v)] for k, v in six.iteritems(flags) if v]
    flags_list = flatten_list(flags_nested)
    return flags_list

def flags_store_iterable(flags):
    '''
    Calls ``flags_store`` for each value within each key in ``flags``.

    e.g. {'--fruits': ['apple', 'pear']} results in ['--fruits', 'apple', '--fruits', 'pear']

    :param dict flags: A dictionary in the form ``{'--flag': value}``, where ``value`` is an iterable.
    :returns: A flat list of strings
    :raises: ``TypeError`` if values are not iterable
    '''
    flags_nested = [[[str(k), str(single_value)] for single_value in v] for k, v in six.iteritems(flags) if v and len(v) > 0]
    flags_list = flatten_list(flatten_list(flags_nested))
    return flags_list

def info(info, quote=True, style=True):
    '''
    Returns the given string ``info`` as a set of echo arguments.

    Optionally provides quoting and terminal styling.

    :param string info: Info string to add to script
    :param bool quote: By default quote given sting in single quotes
    :param bool style: By default add 
    :returns: A flat list of strings
    '''
    info = 'pyqubes|{0}'.format(info)
    if style:
        info = '\e[36m{0}\e[39m'.format(info)
    if quote:
        info = '\'{0}\''.format(info)
    return ['echo', '-e', info]




import pyqubes.compile
import pyqubes.utils
import pyqubes.validate

def qvm_run(vm_name,
            command,
            quiet=False,
            auto=False,
            user='',
            tray=False,
            all_vms=False,
            exclude=[],
            wait=False,
            shutdown=False,
            pause=False,
            unpause=False,
            pass_io=False,
            localcmd='',
            force=False):
    '''
    qvm-run
    '''
    command_args = ["qvm-run", pyqubes.validate.linux_hostname(vm_name), command]
    command_args.extend(pyqubes.compile.flags_boolean({
        '--quiet': quiet,
        '--auto': auto,
        '--tray': tray,
        '--all': all_vms,
        '--wait': wait,
        '--shutdown': shutdown,
        '--pause': pause,
        '--unpause': unpause,
        '--pass-io': pass_io,
        '--force': force
    }))
    command_args.extend(pyqubes.compile.flags_store({
        '--user': pyqubes.validate.linux_username(user) if user else None,
        '--localcmd': localcmd
    }))
    [command_args.extend(pyqubes.compile.flags_store({
        '--exclude': pyqubes.validate.linux_hostname(exclude_single) if exclude_single else None
    })) for exclude_single in exclude]
    return command_args

def qvm_shutdown(vm_name,
            quiet=False,
            force=False,
            wait=False,
            all_vms=False,
            exclude=[]):
    '''
    qvm-shutdown
    '''
    command_args = ["qvm-shutdown", pyqubes.validate.linux_hostname(vm_name)]
    command_args.extend(pyqubes.compile.flags_boolean({
        '--quiet': quiet,
        '--force': force,
        '--wait': wait,
        '--all': all_vms
    }))
    [command_args.extend(pyqubes.compile.flags_store({
        '--exclude': pyqubes.validate.linux_hostname(exclude_single) if exclude_single else None
    })) for exclude_single in exclude]
    return command_args

def qvm_firewall(vm_name,
            list_view=False,
            add_rule='',
            del_rule='',
            set_policy='',
            set_icmp='',
            set_dns='',
            set_yum_proxy='',
            numeric=False):
    '''
    qvm-firewall
    '''
    command_args = ["qvm-firewall", pyqubes.validate.linux_hostname(vm_name)]
    command_args.extend(pyqubes.compile.flags_boolean({
        '--list': list_view,
        '--numeric': numeric
    }))
    command_args.extend(pyqubes.compile.flags_store({
        '--add': add_rule,
        '--del': del_rule,
        '--policy': pyqubes.validate.firewall_policy(set_policy) if set_policy else None,
        '--icmp': pyqubes.validate.firewall_policy(set_icmp) if set_icmp else None,
        '--dns': pyqubes.validate.firewall_policy(set_dns) if set_dns else None,
        '--yum_proxy': pyqubes.validate.firewall_policy(set_yum_proxy) if set_yum_proxy else None
    }))
    return command_args

def qvm_start(vm_name,
            quiet=False,
            no_guid=False,
            console=False,
            dvm=False,
            custom_config=''):
    '''
    qvm-start
    '''
    command_args = ["qvm-start", pyqubes.validate.linux_hostname(vm_name)]
    command_args.extend(pyqubes.compile.flags_boolean({
        '--quiet': quiet,
        '--no-guid': no_guid,
        '--console': console,
        '--dvm': dvm
    }))
    command_args.extend(pyqubes.compile.flags_store({
        '--custom-config': custom_config
    }))
    return command_args

def qvm_clone(vm_name,
            clone_name,
            quiet=False,
            path=''):
    '''
    qvm-clone
    '''
    command_args = ["qvm-clone", pyqubes.validate.linux_hostname(vm_name), pyqubes.validate.linux_hostname(clone_name)]
    command_args.extend(pyqubes.compile.flags_boolean({
        '--quiet': quiet
    }))
    command_args.extend(pyqubes.compile.flags_store({
        '--path': path
    }))
    return command_args

def qvm_create(vm_name,
            template='',
            label='',
            proxy=False,
            net=False,
            hvm=False,
            hvm_template=False,
            root_move_from='',
            root_copy_from='',
            standalone=False,
            mem=0,
            vcpus=0,
            internal=False,
            force_root=False,
            quiet=False):
    '''
    qvm-create
    '''
    command_args = ["qvm-create", pyqubes.validate.linux_hostname(vm_name)]
    command_args.extend(pyqubes.compile.flags_boolean({
        '--proxy': proxy,
        '--net': net,
        '--hvm': hvm,
        '--hvm_template': hvm_template,
        '--standalone': standalone,
        '--internal': internal,
        '--force-root': force_root,
        '--quiet': quiet
    }))
    command_args.extend(pyqubes.compile.flags_store({
        '--template': pyqubes.validate.linux_hostname(template) if template else None,
        '--label': pyqubes.validate.label_color(label) if label else None,
        '--root-move-from': root_move_from,
        '--root-copy-from': root_copy_from,
        '--mem': mem,
        '--vcpus': vcpus
    }))
    return command_args

def qvm_remove(vm_name,
            quiet=False,
            just_db=False,
            force_root=False):
    '''
    qvm-remove
    '''
    command_args = ["qvm-remove", pyqubes.validate.linux_hostname(vm_name)]
    command_args.extend(pyqubes.compile.flags_boolean({
        '--quiet': quiet,
        '--just-db': just_db,
        '--force-root': force_root
    }))
    return command_args




#! /usr/bin/env python
'''
The enact module contains functions that act on a list of command-line arguments,

The two most important ones are:
* Direct execution with ``call`` (proactive mode)
* Echoing an execution-ready script with ``echo`` (reactive mode)
'''
import os
import subprocess
import sys

import six


def echo(args, file=None):
    '''
    Echo a list of arguments (as given to ``subprocess.call``) to the given stream.

    This defaults to ``stdout``, but can be changed to any stream-like object such as a file handle.

    :param args: A string or list of strings
    :param file: A file-like object to stream output to. Defaults to ``sys.stdout``
    '''
    if file is None:
        file = sys.stdout

    if isinstance(args, six.string_types + (six.text_type,)):
        args = [args]
    
    six.print_(*args, file=file, flush=True)

def call(args, **kwargs):
    '''
    Thin wrapper for builtin subprocess.call
    '''
    return subprocess.check_call(args, **kwargs)

def call_quiet(args):
    '''
    Uses the ``call`` function, but throws away stdout and stderr.

    Should be used for internal unit tests wherever possible.
    '''
    with open(os.devnull, 'w') as null_file:
        return call(args, stdout=null_file, stderr=null_file)




#!/usr/bin/env python
'''
Validate functions for pyqubes

These will return the original value if validation passes.
Otherwise, ``ValueError`` will be raised
'''

import re

import pyqubes.constants


# TODO replace with config (regex, lambda, error message)
def linux_username(username):
    '''
    Linux usernames are recommended to match ``^[a-z_][a-z0-9_-]*\$?$``

    :param string username: Username to check
    :returns: ``username`` if valid, else ``ValueError``
    '''
    okay_pattern = True if re.match(r'^[a-z_][a-z0-9_-]*\$?$', username) else False
    okay_length = len(username) <= 32 and len(username) >= 0
    valid_username = okay_length and okay_pattern
    if not valid_username:
        raise ValueError('Invalid linux username: {0}'.format(username))
    else:
        return username

def linux_hostname(hostname):
    '''
    Linux hostnames are recommended to match ``^[a-zA-Z0-9-]+(\.[a-zA-Z0-9-]+)*$``

    :param string hostname: Hostname to check
    :returns: ``hostname`` if valid, else ``ValueError``
    '''
    okay_pattern = True if re.match(r'^[a-zA-Z0-9-]+(\.[a-zA-Z0-9-]+)*$', hostname) else False
    okay_length = len(hostname) <= 253 and len(hostname) >= 0
    valid_hostname = okay_length and okay_pattern
    if not valid_hostname:
        raise ValueError('Invalid linux hostname: {0}'.format(hostname))
    else:
        return hostname

def firewall_policy(policy):
    '''
    qvm-firewall policy string should match ``^(allow|deny)$``

    :param string policy: Policy string to check
    :returns: ``policy`` if valid, else ``ValueError``
    '''
    if policy not in pyqubes.constants.FIREWALL_POLICY_ALL:
        raise ValueError('Invalid qvm-firewall policy: {0}'.format(policy))
    else:
        return policy

def label_color(color):
    '''
    VM label color should be one of:
    * red
    * orange
    * yellow
    * green
    * blue
    * purple
    * black
    * gray

    :param string color: Label color string to check
    :returns: ``color`` if valid, else ``ValueError``
    '''
    if color not in pyqubes.constants.LABEL_COLOR_ALL:
        raise ValueError('Invalid label color: {0}'.format(color))
    else:
        return color




# Operating systems

FEDORA = "fedora"
FEDORA_23 = "fedora-23"

DEBIAN = "debian"
DEBIAN_8 = "debian-8"

FEDORA_ALL = (FEDORA, FEDORA_23)
DEBIAN_ALL = (DEBIAN, DEBIAN_8)
OPERATING_SYSTEM_ALL = FEDORA_ALL + DEBIAN_ALL


# Firewall policy options

ALLOW = 'allow'
DENY = 'deny'
FIREWALL_POLICY_ALL = (ALLOW, DENY)


# VM label colors

RED = 'red'
ORANGE = 'orange'
YELLOW = 'yellow'
GREEN = 'green'
BLUE = 'blue'
PURPLE = 'purple'
BLACK = 'black'
GRAY = 'gray'
LABEL_COLOR_ALL = (RED, ORANGE, YELLOW, GREEN, BLUE, PURPLE, BLACK, GRAY)




#!/usr/bin/env python
'''
Utility functions for pyqubes

Utilities have no dependencies.
'''

import logging

def object_fullname(obj):
    '''Returns the full absolute name of the object provided'''
    fullname = obj.__module__ + "." + obj.__class__.__name__
    return fullname

def object_logger(obj):
    '''
    Returns a correctly named logger for the given object.
    
    Call as self.logger = object_logger(self)
    '''
    fullname = object_fullname(obj)
    logger = logging.getLogger(fullname)
    return logger

def flatten_list(nested_list):
    '''
    Flatten a nested list one level.

    :param list nested_list: A nested list
    :rtype: list
    '''
    flat_list = [item for sublist in nested_list for item in sublist]
    return flat_list

def assert_list_items_equal_in_nested(actual_nested, expected_list):
    '''
    Assert that the given ``expected_list`` matches one pf the lists within ``actual_nested``,
    using the comparison ``sorted(list)``

    :param list actual_nested: A list of lists (usually generated by test)
    :param list expected_list: A list of expected values
    '''
    expected_list_sorted = sorted(expected_list)
    equal_list = [expected_list_sorted == sorted(actual_list) for actual_list in actual_nested]
    if True not in equal_list:
        error_message = 'List not found in nested list.\nList: {0}\nNested list: {1}'.format(expected_list, actual_nested)
        raise AssertionError(error_message)





#!/usr/bin/env python
'''
Python objects representing QubesOS objects
'''

import pyqubes.constants
import pyqubes.enact
import pyqubes.qvm
import pyqubes.validate


class VMMagicFirewall(object):
    '''
    Helper class for opening and closing the VM fireall automatically
    
    :param vm: A ``VM`` instance.
    '''
    def __init__(self, vm):
        self.vm = vm

    def __enter__(self):
        self.vm.firewall_open()

    def __exit__(self, type, value, traceback):
        self.vm.firewall_close()

class VMMagicAnimate(object):
    '''
    Helper class for starting and shutting down the VM automatically
    
    :param vm: A ``VM`` instance.
    '''
    def __init__(self, vm):
        self.vm = vm

    def __enter__(self):
        self.vm.start()

    def __exit__(self, type, value, traceback):
        self.vm.shutdown()

class VM(object):
    '''
    The VM object represants a QubesOS VM. Its methods are common accross
    both AppVMs and TemplateVMs.

    VM should not be instanciated directly - use TemplateVM or AppVM.

    By default, all VMs are Fedora 23 based. Other values are listed in ``pyqubes.constants``
    '''
    def __init__(self, name, proactive=False, operating_system=pyqubes.constants.FEDORA_23):
        self.name = pyqubes.validate.linux_hostname(name)
        self.proactive = proactive
        self.operating_system = operating_system
        
        self.enact_function = pyqubes.enact.call if proactive else pyqubes.enact.echo
        
        self.animate = VMMagicAnimate(self)
        self.internet = VMMagicFirewall(self)
    
    def enact(self, args):
        '''
        Enact a list of command arguments using the VM's ``enact_function``

        Any one of the functions in ``pyqubes.qubes``,``pyqubes.qubesdb``
        or ``pyqubes.qvm`` will return arguments in the correct format.
        '''
        return self.enact_function(args)

    def info(self, info):
        '''
        Echo information from pyqubes using the VM's enact method
        '''
        info = '{0}|{1}'.format(self.name, info)
        return self.enact(pyqubes.compile.info(info))

    # Direct command bindings

    def run(self, command, quote=True, **kwargs):
        '''
        Run a command on the VM.

        Please note:
        * ``--pass-io`` is always set, to run commands synchronously
        * Commands are automatically encapsulated in single quotes::

            vm = TemplateVM('spam')
            vm.run('echo "foo bar"')
            # produces: qvm-run spam 'echo "foo bar"'
        
        :param bool quote: By default command is single quoted - set ``False`` to disable
        '''
        if quote:
            command = "'{0}'".format(command)
        kwargs.update({
            'pass_io': True
        })
        self.info('Running command {0}'.format(command))
        return self.enact(pyqubes.qvm.qvm_run(self.name, command, **kwargs))

    def shutdown(self, **kwargs):
        '''
        Shutdown the 
        
        Please note:
        * ``--wait`` is always set, to run commands synchronously
        '''
        kwargs.update({
            'wait': True
        })
        self.info('Shutting down')
        return self.enact(pyqubes.qvm.qvm_shutdown(self.name, **kwargs))

    def start(self, **kwargs):
        '''
        Start the VM explicitly.

        In most cases you should use``with vm.animate``::

            vm = TemplateVM('foo')
            # Template is not started on instanciation
            with vm.animate:
                # Template is now running
                vm.update()
            # VM is shut down automatically
        
        '''
        self.info('Starting')
        return self.enact(pyqubes.qvm.qvm_start(self.name, **kwargs))

    def firewall(self, **kwargs):
        '''
        Edit the VM firewall
        '''
        self.info('Adjusting firewall')
        return self.enact(pyqubes.qvm.qvm_firewall(self.name, **kwargs))

    def remove(self, **kwargs):
        '''
        Remove the VM
        '''
        self.info('Removing')
        return self.enact(pyqubes.qvm.qvm_remove(self.name, **kwargs))

    # Helper functions

    def firewall_open(self):
        '''
        Can be explicity called to open the VM firewall to 'allow'.

        In most cases you should use``with vm.internet``::

            vm = TemplateVM('foo')
            with vm.animate:
                # Templates are offline be default
                with vm.internet:
                    # Template now has unrestricted internet access
                    vm.run('curl http://ipecho.net/plain')
                # Firewall is restored automatically
                # This will now fail
                vm.run('curl http://ipecho.net/plain')
        
        '''
        self.info('Opening firewall')
        return self.enact(pyqubes.qvm.qvm_firewall(self.name, set_policy='allow'))

    def firewall_close(self):
        '''
        Can be explicity called to close the VM firewall to 'deny'.
        '''
        self.info('Closing firewall')
        return self.enact(pyqubes.qvm.qvm_firewall(self.name, set_policy='deny'))

# TODO: Could TemplateVM and AppVM have metaclassed magic methods?
class TemplateVM(VM):
    '''
    TemplateVM - for installing apps
    '''
    def __init__(self, *args, **kwargs):
        super(TemplateVM, self).__init__(*args, **kwargs)

    def clone(self, clone_name, **kwargs):
        '''
        Clone the TemplateVM and return a new TemplateVM

        :param string clone_name: Name of the new VM
        :returns: The new ``TemplateVM`` instance
        '''
        self.info('Cloning to \'{0}\''.format(clone_name))
        self.enact(pyqubes.qvm.qvm_clone(self.name, clone_name, **kwargs))
        return TemplateVM(clone_name, proactive=self.proactive, operating_system=self.operating_system)

    def create_app(self, app_name, **kwargs):
        '''
        Create and return a new AppVM based on the TemplateVM.
        
        Please note:
        * If ``label`` is not set, it will default to ``red``

        :param string app_name: Name of the new VM
        :returns: The new ``AppVM`` instance
        '''
        kwargs.update({
            'template': self.name,
            'label': kwargs['label'] if kwargs.get('label', None) else pyqubes.constants.RED
        })
        self.info('Creating app \'{0}\''.format(app_name))
        self.enact(pyqubes.qvm.qvm_create(app_name, **kwargs))
        return AppVM(app_name, proactive=self.proactive, operating_system=self.operating_system)
    
    def update(self):
        '''
        Smartly runs the relevant package manager updates for the TemplateVM
        '''
        if self.operating_system in pyqubes.constants.FEDORA_ALL:
            update_command = "sudo dnf -y upgrade --refresh"
        elif self.operating_system in pyqubes.constants.DEBIAN_ALL:
            update_command = "sudo apt-get update && sudo apt-get -y dist-upgrade"
        else:
            raise ValueError("Could not update TemplateVM '{0}': Unknown OS '{1}'".format(self.name, self.operating_system))
        self.info('Updating')
        return self.run(update_command)
    
class AppVM(VM):
    '''
    AppVM - for running apps
    '''
    def __init__(self, *args, **kwargs):
        super(AppVM, self).__init__(*args, **kwargs)




#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name='pyqubes',
    version='0.0.3',
    license='Apache License 2.0',
    url='https://github.com/tommilligan/pyqubes/',
    author='Tom Milligan',
    author_email='code@tommilligan.net',
    description="QubesOS dom0 automation in Python",
    keywords='qubes qubesos QubesOS wrapper recipe dom0 vm templatevm appvm',
    packages=find_packages(exclude=['tests']),
    classifiers=[
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Natural Language :: English',
        'Topic :: Desktop Environment',
        'Topic :: System :: Installation/Setup'
    ],
    zip_safe=False,
    platforms='any',
    install_requires=[
        'six >= 1.10.0'
    ],
    tests_require=['nose2 >= 0.6.5'],
    test_suite='nose2.collector.collector',
    # Install these with "pip install -e isoprene_pumpjack[dev]
    extras_require={
        'dev': [
            'sphinx >= 1.5.3',
            'sphinx-argparse >= 0.1.17',
            'sphinx_rtd_theme >= 0.1.9',
            'codeclimate-test-reporter >= 0.2.1',
            'cov-core >= 1.15.0',
            'nose2 >= 0.6.5',
            'mock >= 2.0.0'
        ]
    },
    entry_points={
        'console_scripts': [
        ]
    },
)
